@@ -1,8 +1,9 @@
 #!/usr/bin/python
 # -*- coding: utf-8 -*-
 
-"""Checkm class library docs TODO
+"""Checksumming convenience classes
 
+TODO! Sorry!
 
                 [@]SourceFileOrURL  Alg     Digest  Length   ModTime   TargetFileOrURL
 TOKEN NUMBER:    1                  2       3       4        5         6
@@ -38,18 +39,41 @@ logging.basicConfig(level=logging.INFO)
 logger = logging.getLogger('checkm')
 
 class NotFound(Exception):
-    """The directory was not found, or is not accessible."""
+    """The item or directory was either not found, or not accessible."""
     def __init__(self, *arg, **kw):
+        """
+        FIXME
+        @param *arg:
+        @type *arg:
+        @param **kw:
+        @type **kw:
+        """
         self.context = (arg, kw)
     def __repr__(self):
+        """
+        FIXME
+        """
+        return self.context.__str__()
+    def __str__(self):
+        """
+        FIXME
+        """
         return self.context.__str__()
 
 class CheckmReporter(object):
     COLUMN_NAMES = [u'# [@]SourceFileOrURL',u'Alg',u'Digest',u'Length',u'ModTime']
     def __init__(self):
+        """
+        FIXME
+        """
         self.scanner = CheckmScanner()
 
     def _get_max_len(self, report):
+        """
+        FIXME
+        @param report:
+        @type report:
+        """
         cols = defaultdict(lambda : 0)
         for line in report:
             for index in xrange(len(line)):
@@ -58,6 +82,13 @@ class CheckmReporter(object):
         return cols
 
     def _space_line(self, line, col_maxes):
+        """
+        FIXME
+        @param line:
+        @type line:
+        @param col_maxes:
+        @type col_maxes:
+        """
         spaced_line = []
         for index in xrange(len(line)):
             spaced_line.append(line[index])
@@ -66,6 +97,19 @@ class CheckmReporter(object):
         return u"".join(spaced_line)
 
     def create_bagit_manifest(self, scan_directory, algorithm, recursive=False, delimiter = "  ", filename=None):
+        """
+        FIXME
+        @param scan_directory:
+        @type scan_directory:
+        @param algorithm:
+        @type algorithm:
+        @param recursive=False:
+        @type recursive=False:
+        @param delimiter:
+        @type delimiter:
+        @param filename=None:
+        @type filename=None:
+        """
         if not filename:
             filename = "manifest-%s.txt" % algorithm
         logger.info("Creating bagit manifest file(%s) for dir(%s) with Alg:%s" % (filename,
@@ -84,27 +128,69 @@ class CheckmReporter(object):
                         output.write("%s%s%s\n" % (line[2], delimiter, line[0]))
                 output.write("\n")
         return filename
+        
+    def create_multilevel_checkm(self, top_directory, algorithm, checkm_filename, columns=3):
+        logger.info("Creating multilevel checkm files '(%s)' from top level directory(%s) with Alg:%s and columns:%s" % (checkm_filename, top_directory, algorithm, columns))
+        if not os.path.isdir(top_directory):
+            raise NotFound(top_directory=top_directory)
+        # Gather list of directories to scan
+        # And their subdirectories
+        # bottom up!
+        dirs = dict([(root, dirnames) for (root, dirnames, _) in os.walk(top_directory, topdown=False)])
+        # per directory
+        for dirname in dirs:
+            with codecs.open(os.path.join(dirname, checkm_filename), encoding='utf-8', mode="w") as output:
+                self.create_checkm_file(dirname, 
+                                        algorithm, 
+                                        os.path.join(dirname, checkm_filename), 
+                                        recursive=False,
+                                        columns=columns,
+                                        checkm_file=output)
+                subdir_report = []
+                for subdir in dirs[dirname]:
+                    try:
+                        line = self.scanner.scan_path(os.path.join(subdir, checkm_filename), algorithm, columns)
+                        line[0] = '@%s' % (line[0])
+                        subdir_report.append(line)
+                    except Exception, e:
+                        print "Fail! %s" % e
+                col_maxes = self._get_max_len(subdir_report)
+                for line in subdir_report:
+                    output.write('%s\n' % (self._space_line(line, col_maxes)))
+                output.write('\n')
 
-    def create_checkm_file(self, scan_directory, algorithm, checkm_filename, recursive=False, columns=3):
-        logger.info("Creating checkm file(%s) for dir(%s) with Alg:%s and columns: %s" % (checkm_filename,
+    def create_checkm_file(self, scan_directory, algorithm, checkm_filename, recursive=False, columns=3, checkm_file=None):
+        logger.info("Creating checkm file for dir(%s) with Alg:%s and columns: %s" % (
                                                                                           scan_directory,
                                                                                           algorithm, columns))
         report = self.scanner.scan_directory(scan_directory, algorithm, recursive=recursive, columns=columns)
         col_maxes = self._get_max_len(report)
-        if hasattr(checkm_filename, 'write'):
-            checkm_filename.write("%s \n" % (self._space_line(CheckmReporter.COLUMN_NAMES[:columns], col_maxes)))
+        if checkm_file != None and hasattr(checkm_file, 'write'):
+            checkm_file.write("%s \n" % (self._space_line(CheckmReporter.COLUMN_NAMES[:columns], col_maxes)))
             for line in report:
-                checkm_filename.write("%s\n" % (self._space_line(line, col_maxes)))
-            checkm_filename.write("\n")
-            return checkm_filename
+                logger.info("Checking that the scanned file is not the empty checkm file")
+                logger.info("Scanned file: %s  vs  Checkm filename: %s" % (line[0], checkm_filename))
+                if line[0] != checkm_filename:
+                    checkm_file.write("%s\n" % (self._space_line(line, col_maxes)))
+                else:
+                    logger.info("MATCH! - scan line ignored")
+            return checkm_file
         else:
             with codecs.open(checkm_filename, encoding='utf-8', mode="w") as output:
                 output.write("%s \n" % (self._space_line(CheckmReporter.COLUMN_NAMES[:columns], col_maxes)))
                 for line in report:
-                    output.write("%s\n" % (self._space_line(line, col_maxes)))
+                    if line[0] != os.path.join(scan_directory, checkm_filename):
+                        output.write("%s\n" % (self._space_line(line, col_maxes)))
                 output.write("\n")
 
     def check_bagit_hashes(self, bagit_filename, algorithm=None):
+        """
+        FIXME
+        @param bagit_filename:
+        @type bagit_filename:
+        @param algorithm=None:
+        @type algorithm=None:
+        """
         logger.info("Checking files against '%s' bagit manifest" % bagit_filename)
         if algorithm == None:
             if hasattr(bagit_filename, 'read'):
@@ -132,43 +218,92 @@ class CheckmReporter(object):
                     results['fail'][row[1]] = (row, scan_row)
         return results
 
-    def check_checkm_hashes(self, scan_directory, checkm_filename):
+    def check_checkm_hashes(self, scan_directory, checkm_filename, ignore_multilevel=True):
+        """
+        FIXME
+        @param scan_directory:
+        @type scan_directory:
+        @param checkm_filename:
+        @type checkm_filename:
+        """
+        def _check_files_against_parser(parser):
+            scanner = CheckmScanner()
+            results = {'pass':[], 'fail':{}, 'include':[]}
+            for row in parser:
+                if row:
+                    try:
+                        if row[0].startswith('@'):
+                            row[0] = row[0][1:]
+                            results['include'].append(row[0])
+                        scan_row = scanner.scan_path(row[0], row[1], len(row))
+                        if row != scan_row:
+                            logger.info("Failed original: %s" % row)
+                            logger.info("Current scan: %s" % scan_row)
+                            results['fail'][row[0]] = (row, scan_row)
+                        else:
+                            results['pass'].append(row[0])
+                    except NotFound:
+                        scan_row = "File not found"
+                        logger.info("Failed original: %s" % row)
+                        logger.info("But file not found at this path.")
+                        results['fail'][row[0]] = (row, scan_row)
+            return results
+        
         logger.info("Checking files against %s checkm manifest" % checkm_filename)
         parser = CheckmParser(checkm_filename)
-        scanner = CheckmScanner()
-        results = {'pass':[], 'fail':{}}
-        for row in parser:
-            if row:
-                try:
-                    scan_row = scanner.scan_path(row[0], row[1], len(row))
-                    if row != scan_row:
-                        logger.info("Failed original: %s" % row)
-                        logger.info("Current scan: %s" % scan_row)
-                        results['fail'][row[0]] = (row, scan_row)
-                    else:
-                        results['pass'].append(row[0])
-                except NotFound:
-                    scan_row = "File not found"
-                    logger.info("Failed original: %s" % row)
-                    logger.info("But file not found at this path.")
-                    results['fail'][row[0]] = (row, scan_row)
-        return results
+        results = _check_files_against_parser(parser)
+        if ignore_multilevel:
+            return results
+        else:
+            # shallow copy of the include list, as we will be pop'ing off items
+            checkm_list = results['include'][:]
+            while checkm_list:
+                checkm_file = checkm_list.pop()
+                parser = CheckmParser(checkm_file)
+                additional_results = _check_files_against_parser(parser)
+                # Add to the passes
+                results['pass'].extend(additional_results['pass'])
+                # add to the overall list of 
+                results['include'].extend(additional_results['include'])
+                checkm_list.extend(additional_results['include'])
+                # add to the fail dict
+                results['fail'].update(additional_results['fail'])
+            return results
 
 class BagitParser(object):
     def __init__(self, bagit_file=None):
+        """
+        FIXME
+        @param bagit_file=None:
+        @type bagit_file=None:
+        """
         self.status = False
         self.lines = []
         if bagit_file:
             self.parse(bagit_file)
 
     def __iter__(self):
+        """
+        FIXME
+        """
         class Bagit_iter:
             def __init__(self, lines):
+                """
+                FIXME
+                @param lines:
+                @type lines:
+                """
                 self.lines = lines
                 self.last = 0
             def __iter__(self):
+                """
+                FIXME
+                """
                 return self
             def next(self):
+                """
+                FIXME
+                """
                 if self.last >= len(self.lines):         # threshhold terminator
                     raise StopIteration
                 elif len(self.lines) == 0:
@@ -179,6 +314,11 @@ class BagitParser(object):
         return Bagit_iter(self.lines)
     
     def parse(self, fileobj):
+        """
+        FIXME
+        @param fileobj:
+        @type fileobj:
+        """
         if not hasattr(fileobj, "read"):
             with codecs.open(fileobj, encoding='utf-8', mode="r") as check_fh:
                 self._parse_lines(check_fh)
@@ -187,9 +327,19 @@ class BagitParser(object):
         return self.lines
 
     def _parse_lines(self, fh):
+        """
+        FIXME
+        @param fh:
+        @type fh:
+        """
         self.lines = [] # clear the deck
         line_buffer = ""
         def _parse_line(line):
+            """
+            FIXME
+            @param line:
+            @type line:
+            """
             if not line.startswith('#'):
                 tokens = filter(lambda x: x, re.split("\s+", line, 1)) # 2 columns
                 logger.info(tokens)
@@ -211,19 +361,38 @@ class BagitParser(object):
 
 class CheckmParser(object):
     def __init__(self, checkm_file=None):
+        """
+        FIXME
+        @param checkm_file=None:
+        @type checkm_file=None:
+        """
         self.status = False
         self.lines = []
         if checkm_file:
             self.parse(checkm_file)
     
     def __iter__(self):
+        """
+        FIXME
+        """
         class Checkm_iter:
             def __init__(self, lines):
+                """
+                FIXME
+                @param lines:
+                @type lines:
+                """
                 self.lines = lines
                 self.last = 0
             def __iter__(self):
+                """
+                FIXME
+                """
                 return self
             def next(self):
+                """
+                FIXME
+                """
                 if self.last >= len(self.lines):         # threshhold terminator
                     raise StopIteration
                 elif len(self.lines) == 0:
@@ -234,22 +403,39 @@ class CheckmParser(object):
         return Checkm_iter(self.lines)
 
     def parse(self, checkm_file):
+        """
+        FIXME
+        @param checkm_file:
+        @type checkm_file:
+        """
         if not hasattr(checkm_file, "readline"):
-            with codecs.open(checkm_file, encoding='utf-8', mode="r") as check_fh:
-                self._parse_lines(check_fh)
+            if os.path.isfile(checkm_file):
+                with codecs.open(checkm_file, encoding='utf-8', mode="r") as check_fh:
+                    self._parse_lines(check_fh)
+            else:
+                raise NotFound(checkm_file=checkm_file)
         else:
             self._parse_lines(checkm_file)
         return self.lines
 
     def _parse_lines(self, fh):
+        """
+        FIXME
+        @param fh:
+        @type fh:
+        """
         self.lines = [] # clear the deck
         line_buffer = ""
         def _parse_line(line):
+            """
+            FIXME
+            @param line:
+            @type line:
+            """
             if not line.startswith('#'):
                 tokens = filter(lambda x: x, re.split("\s+", line, 5)) # 6 column max defn == 5 splits
                 logger.info(tokens)
                 if tokens:
-                    #self.lines.append(dict([(index, tokens[index]) for index in xrange(len(tokens))]))
                     self.lines.append(tokens)
 
         for chunk in fh.read(0x1000):
@@ -266,6 +452,15 @@ class CheckmParser(object):
 class CheckmScanner(object):
     HASHTYPES = ['md5', 'sha1', 'sha224','sha256','sha384','sha512']
     def scan_local(self, directory_path, algorithm, columns=3):
+        """
+        FIXME
+        @param directory_path:
+        @type directory_path:
+        @param algorithm:
+        @type algorithm:
+        @param columns=3:
+        @type columns=3:
+        """
         report = []
         for item in os.listdir(directory_path):
             item_path = os.path.join(directory_path, item)
@@ -273,6 +468,15 @@ class CheckmScanner(object):
         return report
 
     def scan_tree(self, directory_path, algorithm, columns):
+        """
+        FIXME
+        @param directory_path:
+        @type directory_path:
+        @param algorithm:
+        @type algorithm:
+        @param columns:
+        @type columns:
+        """
         report = []
         if os.path.exists(directory_path):
             for (dirpath, dirnames, filenames) in os.walk(directory_path):
@@ -283,6 +487,15 @@ class CheckmScanner(object):
             raise NotFound(directory_path=directory_path, recursive=True)
 
     def scan_path(self, item_path, algorithm, columns):
+        """
+        FIXME
+        @param item_path:
+        @type item_path:
+        @param algorithm:
+        @type algorithm:
+        @param columns:
+        @type columns:
+        """
         if columns<3 or not isinstance(columns, int):
             columns = 3
         try:
@@ -319,6 +532,17 @@ class CheckmScanner(object):
             raise ValueError("This tool cannot perform hashtype %s" % algorithm)
         
     def scan_directory(self, directory_path, algorithm, recursive=False, columns=3):
+        """
+        FIXME
+        @param directory_path:
+        @type directory_path:
+        @param algorithm:
+        @type algorithm:
+        @param recursive=False:
+        @type recursive=False:
+        @param columns=3:
+        @type columns=3:
+        """
         if os.path.exists(directory_path):
             if recursive:
                 return self.scan_tree(directory_path, algorithm, columns)
